// BadFrames plugin for Avisynth - replace bad frames by neighbors or blend
// Copyright(c)2006 Fizick

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .

#define WIN32_LEAN_AND_MEAN

#include "windows.h"
#include "avisynth.h"
#include "badframes.h"
#include <stdio.h>

AVSValue __cdecl Create_BadFrames(AVSValue args, void* user_data, IScriptEnvironment* env)
{
	return new BadFrames(
		args[0].AsClip(),
      args[1],
      args[2].AsBool(true),
      args[3].AsString(""),
		env);
}


extern "C" __declspec(dllexport) const char* __stdcall
AvisynthPluginInit2(IScriptEnvironment* env) {

   env->AddFunction("BadFrames", "ci*[blend]b[file]s", Create_BadFrames, 0);
	return("BadFrames");
}



BadFrames::BadFrames(PClip _child, AVSValue frames, bool _blend, const char *_filename, IScriptEnvironment *env) :
GenericVideoFilter(_child)
{

	blend = _blend;

	FILE *file;

	filename = _filename;
	char line[256];

   nFrames = vi.num_frames;
   AllFrames = new BYTE[nFrames];

   memset(AllFrames, GOODFRAME, nFrames);

   AllFrames[0] = NEWSCENEFRAME; // first

	int nBadFrames;

    if (filename != "") // // use list from file
    {
		file = fopen(filename,"rt");
		if (file == NULL) env->ThrowError("BadFrames: input file is not found!");
		int nf;
        while (!feof(file)){
            fgets(line,254,file);
       		sscanf(line,"%d",&nf);
            if (nf >= 0) {
               if (nf < nFrames)
                    AllFrames[nf] = BADFRAME;
            }
            else {
                if (-nf < nFrames)
                    AllFrames[-nf] = NEWSCENEFRAME;
            }
        }
    }

    // additionally use list from parameters
       nBadFrames = frames.ArraySize();
       for (int i=0; i<nBadFrames; i++)
       {
           int nf = frames[i].AsInt();
           if (nf >= 0)
                AllFrames[nf] = BADFRAME;
           else
               AllFrames[-nf] = NEWSCENEFRAME;
       }

}

BadFrames::~BadFrames()
{
	delete AllFrames;
}

void InterpolatePlane(BYTE *pDst, int dst_pitch, const BYTE *pPrev, int prev_pitch, const BYTE *pNext, int next_pitch, int width, int height, int pfraction256)
{
	for (int h=0; h<height; h++)
	{
		for (int w=0; w<width; w++)
		{
			pDst[w] = (pPrev[w]*pfraction256 + pNext[w]*(256-pfraction256))>>8;
		}
		pPrev += prev_pitch;
		pNext += next_pitch;
		pDst += dst_pitch;
	}
}


PVideoFrame __stdcall BadFrames::GetFrame(int n, IScriptEnvironment* env)
{
   int prev_pitch, next_pitch, dst_pitch;
   int width, height;
   const BYTE * pPrev;
   const BYTE * pNext;
   BYTE * pDst;
   int plane;
   int pfraction256;

   PVideoFrame src = child->GetFrame(n, env);

   if (AllFrames[n] != BADFRAME)
	   return src;

   // get good prev
   int nprev = n;
   while (AllFrames[nprev] == BADFRAME && nprev>0)
	   nprev--;
   // get good next
   int nnext = n;
   while (AllFrames[nnext] == BADFRAME && nnext<nFrames-1)
	   nnext++;

	PVideoFrame prev = child->GetFrame(nprev, env);
	PVideoFrame next = child->GetFrame(nnext, env);
   PVideoFrame dst = env->NewVideoFrame(vi);

	if ( (nprev == n || AllFrames[nprev]==BADFRAME) && (nnext>n && AllFrames[nnext]==GOODFRAME) )
		return next;
	else
	if ( (nnext == n || AllFrames[nnext]!=GOODFRAME) && (nprev<n && AllFrames[nprev]!=BADFRAME) )
		return prev;
	else
	if ( (nprev==n || AllFrames[nprev]==BADFRAME) && (nnext==n || AllFrames[nnext]!=GOODFRAME) )
		return src; // very bad case, no any good neighbor

	if (!blend)
		return prev;

	   // interpolate (blend)
	   pfraction256 = ((nnext-n)*256)/(nnext-nprev);

	   if(vi.IsPlanar())
	   {
			plane=PLANAR_Y;
			  pPrev = prev->GetReadPtr(plane);
			  prev_pitch = prev->GetPitch(plane);
			  pNext = next->GetReadPtr(plane);
			  next_pitch = next->GetPitch(plane);
			  pDst = dst->GetWritePtr(plane);
			  dst_pitch = dst->GetPitch(plane);
			  width = dst->GetRowSize(plane);
			  height = dst->GetHeight(plane);
			  InterpolatePlane(pDst, dst_pitch, pPrev, prev_pitch, pNext, next_pitch, width, height, pfraction256);
			plane=PLANAR_U;
			  pPrev = prev->GetReadPtr(plane);
			  prev_pitch = prev->GetPitch(plane);
			  pNext = next->GetReadPtr(plane);
			  next_pitch = next->GetPitch(plane);
			  pDst = dst->GetWritePtr(plane);
			  dst_pitch = dst->GetPitch(plane);
			  width = dst->GetRowSize(plane);
			  height = dst->GetHeight(plane);
			  InterpolatePlane(pDst, dst_pitch, pPrev, prev_pitch, pNext, next_pitch, width, height, pfraction256);
			plane=PLANAR_V;
			  pPrev = prev->GetReadPtr(plane);
			  prev_pitch = prev->GetPitch(plane);
			  pNext = next->GetReadPtr(plane);
			  next_pitch = next->GetPitch(plane);
			  pDst = dst->GetWritePtr(plane);
			  dst_pitch = dst->GetPitch(plane);
			  width = dst->GetRowSize(plane);
			  height = dst->GetHeight(plane);
			  InterpolatePlane(pDst, dst_pitch, pPrev, prev_pitch, pNext, next_pitch, width, height, pfraction256);
	   }
	   else //yuy2, RGB
	   {
			plane=0;
			  pPrev = prev->GetReadPtr(plane);
			  prev_pitch = prev->GetPitch(plane);
			  pNext = next->GetReadPtr(plane);
			  next_pitch = next->GetPitch(plane);
			  pDst = dst->GetWritePtr(plane);
			  dst_pitch = dst->GetPitch(plane);
			  width = dst->GetRowSize(plane);
			  height = dst->GetHeight(plane);
			  InterpolatePlane(pDst, dst_pitch, pPrev, prev_pitch, pNext, next_pitch, width, height, pfraction256);
	   }

	   return dst;
}
